/*
** iplog_dns.c - iplog DNS cache routines.
** Copyright (C) 1999,2000 Ryan McCabe <odin@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
**
** $Id: iplog_dns.c,v 1.7 2000/01/14 18:47:06 odin Exp $
*/

#include <config.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <signal.h>
#include <ctype.h>
#include <netdb.h>
#include <string.h>
#include <pthread.h>
#include <sys/socket.h>

#include <iplog.h>
#include <iplog_dns.h>
#include <iplog_options.h>

static u_char *add_to_cache(addr_t addr, u_long hash, u_char *, size_t);

static size_t dt_size;
static struct dns_entry *dns_table;

extern u_long flags;

void init_dns_table(size_t tlen) {
	dt_size = tlen;
	dns_table = xcalloc(dt_size, sizeof(struct dns_entry));
}

void expire_dns(void) {
	size_t i;
	struct dns_data *cur;

	for (i = 0 ; i < dt_size ; i++) {
		pthread_mutex_lock(&dns_table[i].lock);
		cur = dns_table[i].head;
		while (cur != NULL) {
			if (time(NULL) >= cur->expire) {
				free(cur->host);
				cur = list_delete(cur, (void **) &dns_table[i].head);
				dns_table[i].cnt--;
			} else
				cur = cur->next;
		}
		pthread_mutex_unlock(&dns_table[i].lock);
	}
}

#ifdef HAVE_PTHREAD_CANCEL
static void dns_cleanup(void *data) {
	struct dns_data *cur = data;

	free(cur->host);
}

void destroy_dns_cache(void) {
	size_t i;

	/*
	** All the threads have been canceled, and the locks could be in any
	** state, just zero the whole table out
	*/
	for (i = 0 ; i < dt_size ; i++) {
		if (dns_table[i].head != NULL)
			list_destroy(dns_table[i].head, dns_cleanup);
		memset(&dns_table[i], 0, sizeof(struct dns_entry));
	}
}
#endif

u_char *get_dns_cache(addr_t addr, u_char *buf, size_t len) {
	u_long hash = DNS_HASH(addr, dt_size);
	bool found = false;
	struct dns_data *cur;

	pthread_mutex_lock(&dns_table[hash].lock);
	for (cur = dns_table[hash].head ; cur != NULL ; cur = cur->next) {
		if (cur->addr == addr) {
			xstrncpy(buf, cur->host, len);
			found = true;
			break;
		}
	}
	pthread_mutex_unlock(&dns_table[hash].lock);

	if (found == false)
		return (add_to_cache(addr, hash, buf, len));

	return (buf);
}

static u_char *add_to_cache(addr_t addr, u_long hash, u_char *buf, size_t len) {
	struct dns_data temp;
	struct in_addr in;

	in.s_addr = addr;
	_host_lookup(&in, buf, len);

	temp.next = NULL;
	temp.addr = addr;
	temp.expire = time(NULL) + DNS_TIMEOUT;

	if (flags & LOG_IP && !isdigit(buf[strlen(buf) - 1])) {
		u_long mlen = strlen(buf) + MAX_IPLEN + 4;
		u_char ibuf[MAX_IPLEN];

		temp.host = xmalloc(mlen);
		inet_ntoa_r(&in, ibuf, sizeof(ibuf));
		snprintf(temp.host, mlen, "%s (%s)", buf, ibuf);

		if (mlen <= len)
			xstrncpy(buf, temp.host, len);
		else
			xstrncpy(buf, ibuf, len);
	} else
		temp.host = xstrdup(buf);

	pthread_mutex_lock(&dns_table[hash].lock);
	if (dns_table[hash].cnt < DNS_MAX_ENT) {
		list_prepend(&temp, (void **) &dns_table[hash].head, sizeof(temp));
		dns_table[hash].cnt++;
	} else {
		u_long old_t = ~0;
		struct dns_data *cur, *oldest = NULL;

		for (cur = dns_table[hash].head ; cur != NULL ; cur = cur->next) {
			if ((u_long) cur->expire <= old_t) {
				old_t = cur->expire;
				oldest = cur;
			}
		}
		list_delete(oldest, (void **) &dns_table[hash].head);
		list_prepend(&temp, (void **) &dns_table[hash].head, sizeof(temp));
	}
	pthread_mutex_unlock(&dns_table[hash].lock);

	return (buf);
}

/* vim:ts=4:sw=8:tw=0 */
