/*
** iplog_scan.c - iplog scan/flood detector.
** Copyright (C) 1999,2000 Ryan McCabe <odin@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
**
** $Id: iplog_scan.c,v 1.11 2000/01/18 02:32:09 odin Exp $
*/

/*
** This should now work when iplog is operating in promiscuous mode.  How
** well it wll work is another question.. This may need to be overhauled,
** depending on how well (or poorly) it works in practice.
*/

#include <config.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <pthread.h>
#include <rwlock.h>

#include <iplog.h>
#include <iplog_scan.h>
#include <iplog_options.h>

#define ENDSCAN \
	"%s mode expired for %s - received a total of %lu packets (%lu bytes)."

#define ENDSCAN_P \
"%s mode to %s expired for %s - received a total of %lu packets (%lu bytes)."

extern u_long flags;

static size_t st_size;
static struct scan_list *scan_table;

static void log_end_scan(const struct scan_data *cur, u_char type);

__inline__ static bool no_res(void) {
	return (0);
}

static const struct scan_info scan[] = {
	{ "TCP: port scan",			tcp_res,	PS_TIMEOUT,		PS_THOLD	},
	{ "TCP: null scan",			tcp_res,	NS_TIMEOUT,		NS_THOLD	},
	{ "TCP: FIN scan",			tcp_res,	FS_TIMEOUT,		FS_THOLD	},
	{ "TCP: Xmas scan",			tcp_res,	XS_TIMEOUT,		XS_THOLD	},
	{ "UDP: scan/flood",		udp_res,	UDP_TIMEOUT,	UDP_THOLD	},
	{ "ICMP/UDP: smurf attack",	no_res,		SMURF_TIMEOUT,	SMURF_THOLD	},
	{ "ICMP: ping flood",		icmp_res,	PING_TIMEOUT,	PING_THOLD	}
};

static void log_end_scan(const struct scan_data *cur, u_char type) {
	u_char buf[MAX_HSTLEN];
	struct in_addr in;

	in.s_addr = cur->src_addr;
	host_lookup(&in, scan[type].resolv(), buf, sizeof(buf));

	if (flags & PROMISC) {
		u_char dbuf[MAX_HSTLEN];

		in.s_addr = cur->dst_addr;
		host_lookup(&in, scan[type].resolv(), dbuf, sizeof(dbuf));

		mysyslog(ENDSCAN_P, scan[type].name, dbuf, buf,
				cur->type[type].count, cur->type[type].bytes);
	} else {
		mysyslog(ENDSCAN, scan[type].name, buf, 
				cur->type[type].count, cur->type[type].bytes);
	}
}

void init_scan_table(size_t tsize) {
	st_size = tsize;
	scan_table = xcalloc(st_size, sizeof(struct scan_list));
}

bool check_scan(u_char type, const struct ip *ip, u_long len) {
	u_long hash, count;
	addr_t addr, daddr;
	struct scan_data *cur;
	struct scan_t *data;
	time_t cur_time;
	bool ret = false, found = false, log_scan = false;

	daddr = ip->ip_dst.s_addr;
	addr = ip->ip_src.s_addr;

	if ((flags & SMURF) && type == SCAN_SMURF)
		addr &= 0xffffff;

	hash = SCAN_HASH(addr, daddr, st_size);

	rwl_readlock(&scan_table[hash].lock);
	for (cur = scan_table[hash].head ; cur != NULL ; cur = cur->next) {
		pthread_mutex_lock(&cur->lock);
		if (cur->src_addr == addr && cur->dst_addr == daddr) {
			found = true;
			data = &cur->type[type];

			time(&cur_time);
			cur->last = cur_time;
			data->expire = cur_time + scan[type].timeout;
			data->bytes += len;

			if (++data->count >= scan[type].threshold) {
				ret = true;
				if (data->logged == false) {
					data->logged = true;
					log_scan = true;
				}
			}

			pthread_mutex_unlock(&cur->lock);
			break;
		}
		pthread_mutex_unlock(&cur->lock);
	}

	count = scan_table[hash].cnt;
	rwl_readunlock(&scan_table[hash].lock);

	if (found == false) {
		struct scan_data *tcur, temp;

		if (count >= SCAN_MAXENT) {
			/*
			** This node is full.  Evict the oldest entry.
			*/
			size_t j;
			u_long oldest_time = ~0;
			struct scan_data *oldest = NULL;

			rwl_readlock(&scan_table[hash].lock);
			for (tcur = scan_table[hash].head ; tcur ; tcur = tcur->next) {
				if ((u_long) tcur->last <= oldest_time) {
					oldest_time = tcur->last;
					oldest = tcur;
				}
			}
			pthread_mutex_lock(&oldest->lock);
			rwl_readunlock(&scan_table[hash].lock);

			for (j = 0 ; j < SCAN_TOTAL ; j++) {
				if (oldest->type[j].count >= scan[j].threshold)
					log_end_scan(oldest, j);
			}

			rwl_writelock(&scan_table[hash].lock);
			pthread_mutex_unlock(&oldest->lock);
			cur = list_delete(oldest, (void **) &scan_table[hash].head);
			scan_table[hash].cnt--;
			rwl_writeunlock(&scan_table[hash].lock);
		}

		memset(&temp, 0, sizeof(temp));
		temp.src_addr = addr;
		temp.dst_addr = daddr;

		rwl_writelock(&scan_table[hash].lock);
		tcur = list_prepend(&temp, (void **) &scan_table[hash].head,
							sizeof(temp));
		scan_table[hash].cnt++;
		rwl_writeunlock(&scan_table[hash].lock);

		pthread_mutex_lock(&tcur->lock);

		data = &tcur->type[type];
		data->count = 1;
		data->bytes = len;

		time(&cur_time);
		data->expire = cur_time + scan[type].timeout;
		tcur->last = cur_time;

		pthread_mutex_unlock(&tcur->lock);
	}

	if (log_scan == true) {
		u_char buf[MAX_HSTLEN];
		struct in_addr in;

		in.s_addr = addr;
		host_lookup(&in, scan[type].resolv(), buf, sizeof(buf));

		if (flags & PROMISC) {
			u_char dbuf[MAX_HSTLEN];

			host_lookup(&ip->ip_dst, scan[type].resolv(), dbuf, sizeof(dbuf));
			mysyslog("%s detected to %s from %s", scan[type].name, dbuf, buf);
		} else
			mysyslog("%s detected from %s",	scan[type].name, buf);

		ret = true;
	}

	return (ret);
}

void expire_scans(void) {
	size_t i, j;
	struct scan_data *cur;
	bool remove_entry;

	for (i = 0 ; i < st_size ; i++) {
		rwl_writelock(&scan_table[i].lock);
		cur = scan_table[i].head;

		for (remove_entry = true ; cur != NULL ;) {
			pthread_mutex_lock(&cur->lock);
			for (j = 0 ; j < SCAN_TOTAL ; j++) {
				if (cur->type[j].count != 0 &&
					time(NULL) >= cur->type[j].expire)
				{
					if (cur->type[j].count >= scan[j].threshold)
						log_end_scan(cur, j);

					memset(&cur->type[j], 0, sizeof(struct scan_t));
				} else if (cur->type[j].count != 0)
					remove_entry = false;
			}
			pthread_mutex_unlock(&cur->lock);

			if (remove_entry == true) {
				cur = list_delete(cur, (void **) &scan_table[i].head);
				scan_table[i].cnt--;
			} else
				cur = cur->next;
		}

		rwl_writeunlock(&scan_table[i].lock);
	}
}

#ifdef HAVE_PTHREAD_CANCEL
void destroy_scan_table(void) {
	size_t i;

	for (i = 0 ; i < st_size ; i++) {
		if (scan_table[i].head != NULL)
			list_destroy(scan_table[i].head, NULL);
		memset(&scan_table[i], 0, sizeof(struct scan_list));
	}
}
#endif

/* vim:ts=4:sw=8:tw=0 */
